from kafkacrypto.exceptions import KafkaCryptoMessageError

class KafkaCryptoMessage(object):
  """Class encapsulating crypto-handled KafkaCrypto messages. If a message is
  nominally ciphertext, some conditions mean it should be decryptable, while
  other conditions mean it will never be decryptable, so we keep track of that.

  Keyword Arguments:
                     msg (bytes): Message to turn into a KafkaCryptoMessage object
             ipt (bool,optional): Is message plaintext, or ciphertext?
                                  Default: plaintext.
            deser (obj,optional): Deserializer that can be used to attempt decryption
                                  if message is not plaintext
            topic (str,optional): Topic from which this message came
     decryptable (bool,optional): If the message is ciphertext, will it ever
                                  be decryptable (if keys become available)?
                                  Default: True.
  """
  #
  # Per instance, defined in init
  #   _msg: message bytes
  #   _ipt: plainext boolean
  # _deser: message deserializer
  # _topic: message topic
  # _decryptable: decryptability boolean
  #
  def __init__(self, msg, ipt=True, deser=None, topic=None, decryptable=True):
    if (not isinstance(msg, (bytes,bytearray))):
      raise KafkaCryptoMessageError("Passed message not bytes-like!")
    if (not isinstance(ipt, (bool))):
      raise KafkaCryptoMessageError("Passed plaintext flag not a boolean!")
    self._msg = msg
    self._ipt = ipt
    self._deser = deser
    self._topic = topic
    self._decryptable = decryptable
    if self._ipt:
      self._deser = None
      self._topic = None
      self._decryptable = True

  def isCleartext(self, retry=True):
    if self._deser != None and self._topic != None and self._decryptable and retry:
      # Attempt decryption
      try:
        kcm = self._deser.deserialize(self._topic, self._msg)
        if kcm._ipt:
          # success
          self._msg = kcm._msg
          self._ipt = True
          self._deser = None
          self._topic = None
          self._decryptable = True
        else:
          # failure, update decryptability status if needed
          self._decryptable = kcm._decryptable
      except:
        pass
    return self._ipt

  def isPossiblyDecryptable(self):
    return self._decryptable

  def getMessage(self, retry=True):
    if (not self.isCleartext(retry=retry)):
      raise KafkaCryptoMessageError("Message not decrypted!")
    return self._msg

  def __bytes__(self):
    return self._msg

  @staticmethod
  def fromEncryptedBytes(rawmsg,deser=None,topic=None,decryptable=True):
    if (not isinstance(rawmsg, (bytes,bytearray))):
      raise KafkaCryptoMessageError("Passed raw message not bytes-like!")
    if (len(rawmsg) == 0):
      # zero length message, never decryptable
      return KafkaCryptoMessage(b'',False,decryptable=False)
    elif not (topic is None) and not (deser is None):
      # message not yet decrypted
      return KafkaCryptoMessage(rawmsg,False,deser=deser,topic=topic,decryptable=decryptable)
    else:
      # without deser and topic, will never be decryptable
      return KafkaCryptoMessage(rawmsg,False,deser=deser,topic=topic,decryptable=False)



import pysodium
import msgpack
from hashlib import sha256
from os import replace, remove, fsync as osfsync, open as osopen, close as osclose
from os.path import dirname, normpath
from shutil import copy, copymode
from base64 import b64encode, b64decode
from binascii import unhexlify, hexlify, Error as binasciiError
from time import time
from kafkacrypto.exceptions import KafkaCryptoUtilError
from traceback import format_exception
import inspect
try:
  from dirhash import dirhash
except:
  def dirhash(*args, **kwargs):
    return 'UNKNOWN (is dirhash installed?)'
from os.path import dirname

# Get a sha256 hash representation of code version
def kafkacrypto_hash_version():
  return dirhash(dirname(__file__), "sha256", match=["*.py"])

# generic msgpack packer
def msgpack_default_pack(x):
  if hasattr(x, 'msgpack_pack') and inspect.isroutine(x.msgpack_pack):
    return x.msgpack_pack()
  else:
    return x

# *nix specific items
try:
    import fcntl

    # We need to be able to exclusively lock a file in the implementation of
    # AtomicWriter here. Takes as input a file descriptor.
    def exclusive_lock(fh):
        fcntl.lockf(fh.fileno(),fcntl.LOCK_EX | fcntl.LOCK_NB) # Locks entire file exclusively
    def exclusive_unlock(fh):
        fcntl.lockf(fh.fileno(),fcntl.LOCK_UN) # Unlock file exclusively

    # See https://developer.apple.com/library/archive/documentation/System/Conceptual/ManPages_iPhoneOS/man2/fsync.2.html
    # Need F_FULLFSYNC on OS X / iOS to actually flush everything
    if hasattr(fcntl, 'F_FULLFSYNC'):
        def osfsync(fd):
            fcntl.fcntl(fd, fcntl.F_FULLFSYNC)
except ImportError:
    pass

# Windows specific items
try:
    import msvcrt
    from os import fstat as osfstat

    # We need to be able to exclusively lock a file in the implementation of
    # AtomicWriter here. Takes as input a file descriptor.
    # Per:
    #   https://bugs.python.org/issue29392
    #   https://docs.microsoft.com/en-us/cpp/c-runtime-library/reference/locking?view=msvc-170
    # We have to find out the size of the file to lock all of it. But for
    # a file being written to, its size might be different at lock time than unlock time.
    # We can, however, lock past the end of file. So lock the maximum amount we can (2^31-1),
    # and unlock the same. We also have to make sure the lock starts from the beginning of the file.
    def exclusive_lock(fh):
        pos = fh.tell()
        fh.seek(0)
        msvcrt.locking(fh.fileno(),msvcrt.LK_NBLCK,2147483647) # Locks entire file exclusively
        fh.seek(pos)
    def exclusive_unlock(fh):
        pos = fh.tell()
        fh.seek(0)
        msvcrt.locking(fh.fileno(),msvcrt.LK_UNLCK,2147483647) # Unlock file exclusively
        fh.seek(pos)
except ImportError:
    pass

# Shim for Python 3.10+ compatibility
def format_exception_shim(exc=None, **kwargs):
    try:
        # Python < 3.10
        return format_exception(etype=type(exc), value=exc, tb=exc.__traceback__, **kwargs)
    except:
        try:
            # Python >= 3.10
            return format_exception(exc, **kwargs)
        except:
            # Catch any errors to avoid management thread dying
            return ["Could not format exception.\n"]

# Useful for log rate limiting
log_limited_map = {}
def log_limited(logfunc, msg, *args, msg_id=None, limit_time=10, **kwargs):
    global log_limited_map
    if msg_id is None:
        msgh = msg
        if isinstance(msgh, (str,)):
            msgh = msgh.encode('utf-8')
        msg_id = sha256(msgh).digest()
    if not (msg_id in log_limited_map) or log_limited_map[msg_id]+limit_time <= time():
        log_limited_map[msg_id] = time()
        logfunc(msg, *args, **kwargs)

def str_shim_eq(v, lit):
  if isinstance(v,(bytes,bytearray)):
    v = v.decode('utf-8')
  if isinstance(lit,(bytes,bytearray)):
    lit = lit.decode('utf-8')
  if v == lit:
    return True
  return False

def str_shim_ne(v, lit):
  if isinstance(v,(bytes,bytearray)):
    v = v.decode('utf-8')
  if isinstance(lit,(bytes,bytearray)):
    lit	= lit.decode('utf-8')
  if v != lit:
    return True
  return False

def str_decode(value, iskey=False):
  if value!=None:
    try:
      value = int(value)
    except ValueError as e:
      try:
        value = float(value)
      except ValueError as e:
        if value.lower() == "true":
          value = True 
        elif value.lower() == "false":
          value = False
        elif value.startswith("base64#"):
          try:
            value = b64decode(value[7:].encode('utf-8'),validate=True)
          except binasciiError:
            value = None
        elif iskey:
          # case insensitive keys
          value = value.lower()
  elif default!=None:
    return default
  return value

def str_encode(value, iskey=False):
  if value!=None:
    if isinstance(value,(int,float,bool)):
      value = str(value)
    if not isinstance(value,(str,)):
      value = 'base64#' + b64encode(value).decode('utf-8')
    elif iskey:
      # case insensitive keys
      value = value.lower()
  return value

class AtomicFile(object):
  """The AtomicFile class instantiates a very simplistic file with atomic semantics.
     Nothing is written until flush is issued, then all or nothing is written).
     Because multiple writes can be bunched and then committed atomically, we are
     required to be the only process reading or writing the file (a more elaborate
     implementation could elide this requirement, but is not needed in KafkaCrypto).

     There is a known race condition where there are brief windows in time (during
     the flush operation) when another process could swoop in and take over access.
     However, this will be "noisy" as then the exlusive lock request at the end
     of the flush will fail. So it achieves the needed guarantee here: at most one
     AtomicFile object is associated with a given file at a time.

     It IS NOT thread safe.
  """
  def __init__(self, file, binary=False, tmpfile=None):
    if tmpfile==None:
      tmpfile = file + ".tmp"
    self.__file = file
    self.__filedir = normpath(dirname(file))
    self.__binary = binary
    if not binary:
      self.__fm = "r+"
    else:
      self.__fm = "rb+"
    self.__tmpfile = tmpfile
    # Readfile is only used for reading. However, to enable exclusive access, we
    # must actually open it for writing too
    self.__readfile = self.internal_open(self.__file, self.__fm)
    # Writefile is only used for writing. However, to keep read/write consistency,
    # we must open it for both reading and writing
    self.__writefile = None

  def internal_open(self, *args, **kwargs):
    rv = open(*args, **kwargs)
    exclusive_lock(rv)
    return rv

  def internal_close(self, ofo):
    exclusive_unlock(ofo)
    return ofo.close()

  def seek(self, *args, **kwargs):
    if self.__writefile != None:
      self.__writefile.seek(*args, **kwargs)
    return self.__readfile.seek(*args, **kwargs)

  def write(self, *args, **kwargs):
    if self.__writefile == None:
      exclusive_unlock(self.__readfile)
      copy(self.__file, self.__tmpfile)
      exclusive_lock(self.__readfile)
      self.__writefile = self.internal_open(self.__tmpfile, self.__fm)
      self.__writefile.seek(self.__readfile.tell(),0)
    return self.__writefile.write(*args, **kwargs)

  def writelines(self, lines):
    if self.__binary:
      combined = b''.join(lines)
    else:
      combined = ''.join(lines)
    return self.write(lines)

  def truncate(self, **kwargs):
    if self.__writefile == None:
      exclusive_unlock(self.__readfile)
      copy(self.__file,	self.__tmpfile)
      exclusive_lock(self.__readfile)
      self.__writefile = self.internal_open(self.__tmpfile, self.__fm)
      self.__writefile.seek(self.__readfile.tell(),0)
    return self.__writefile.truncate(**kwargs)

  def flush_dir(self):
    try:
      fd = osopen(self.__filedir, 0)
      try:
        osfsync(fd)
      finally:
        osclose(fd)
    except PermissionError:
      pass # Ignorable since it just means atomic operation may take longer to complete (but still atomic)
           # Happens on Windows

  def flush(self):
    if self.__writefile != None:
      # flush written data to temporary file
      rv = self.__writefile.flush()
      # preserve current file read/write position to restore later
      pos = self.__writefile.tell()
      # sync and close reader and writer handles
      self.internal_close(self.__readfile)
      osfsync(self.__writefile.fileno())
      self.internal_close(self.__writefile)
      # copy file permissions, etc, to tmpfile, in preparation for atomic replace
      # both files are closed, so no exlusive locks to release first
      copymode(self.__file, self.__tmpfile)
      # the atomic operation
      # atomic on python 3.3+ on POSIX
      # atomic on python 3.3+ on Windows NT if filesystem supports atomic moves
      #   https://social.msdn.microsoft.com/Forums/windowsdesktop/en-US/449bb49d-8acc-48dc-a46f-0760ceddbfc3/movefileexmovefilereplaceexisting-ntfs-same-volume-atomic?forum=windowssdk#a239bc26-eaf0-4920-9f21-440bd2be9cc8
      #   https://github.com/python/cpython/blob/main/Modules/posixmodule.c#L4728
      replace(self.__tmpfile, self.__file)
      # The above is atomic, but not necessarily committed until the containing directory's metadata is updated.
      # (which is not guaranteed on windows NT since MOVEFILE_WRITE_THROUGH is not passed in Pythons replace implementation, and not guaranteed on POSIX either)
      # Issue the directory fsync to commit it here.
      self.flush_dir()
      # reopen for reading, restoring read pointer location
      self.__readfile = self.internal_open(self.__file, self.__fm)
      self.__readfile.seek(pos,0)
      self.__writefile = None
      return rv
    else:
      remove(self.__tmpfile)
      return None

  def close(self):
    self.flush()
    return self.internal_close(self.__readfile)

  def __iter__(self):
    return self.__readfile.__iter__()

  def __next__(self):
    return self.__readfile.__next__()

  def __getattr__(self, name):
    return getattr(self.__readfile,name)

def atomic_open(file, mode='r', **kwargs):
  mode = kwargs.pop('mode', mode)
  if len(kwargs) > 0 or (mode.lower() != 'r+' and mode.lower() != 'rb+'):
    raise ValueError("Not Supported!")
  binary = False
  if mode.lower().find('b') != -1:
    binary = True
  return AtomicFile(file, binary=binary)

class PasswordProvisioner(object):
  """The PasswordProvisioner class instantiates three provisioner keys, and uses them
  to sign signing keys of new producers/consumers.

  Alternative versions using, e.g. security keys or similar, can also be written.

  Keyword Arguments:
  password (str): Password from which to derive the provisioning secret keys.

  """
  # Constrained devices cannot use larger numbers than interactive
  _ops = pysodium.crypto_pwhash_scryptsalsa208sha256_OPSLIMIT_INTERACTIVE
  _mem = pysodium.crypto_pwhash_scryptsalsa208sha256_MEMLIMIT_INTERACTIVE

  def __init__(self, password, rot):
    if (isinstance(password,(str,))):
      password = password.encode('utf-8')
    try:
      rot = unhexlify(rot)
    except:
      pass
    self._salt = {}
    self._salt['producer'] = pysodium.crypto_hash_sha256(b'producer' + rot)[0:pysodium.crypto_pwhash_scryptsalsa208sha256_SALTBYTES]
    self._salt['consumer'] = pysodium.crypto_hash_sha256(b'consumer' + rot)[0:pysodium.crypto_pwhash_scryptsalsa208sha256_SALTBYTES]
    self._salt['prodcon'] = pysodium.crypto_hash_sha256(b'prodcon' + rot)[0:pysodium.crypto_pwhash_scryptsalsa208sha256_SALTBYTES]
    self._seed = {}
    self._pk = {}
    self._sk = {}
    print("")
    print("Deriving provisioning keys with:")
    print("  opsl = ", self._ops)
    print("  meml = ", self._mem)
    for key in self._salt.keys():
      print("  salt (", key, ") = ", hexlify(self._salt[key]))
      self._seed[key] = pysodium.crypto_pwhash_scryptsalsa208sha256(pysodium.crypto_sign_SEEDBYTES, password, self._salt[key], opslimit=self._ops, memlimit=self._mem)
      self._pk[key], self._sk[key] = pysodium.crypto_sign_seed_keypair(self._seed[key])
      print("  Signing Public Key (", key, "): ", hexlify(self._pk[key]))
    print("")

class PasswordROT(object):
  """The PasswordROT class instantiates a root of trust keypair.

  Alternative versions using, e.g. security keys or similar, can also be written.

  Keyword Arguments:
  password (str): Password from which to derive the secret key.
  """
  # Constrained devices cannot use larger numbers than interactive
  _ops = pysodium.crypto_pwhash_scryptsalsa208sha256_OPSLIMIT_INTERACTIVE
  _mem = pysodium.crypto_pwhash_scryptsalsa208sha256_MEMLIMIT_INTERACTIVE

  def __init__(self, password):
    if (isinstance(password,(str,))):
      password = password.encode('utf-8')
    self._salt = pysodium.crypto_hash_sha256(b'Root of Trust' + password)[0:pysodium.crypto_pwhash_scryptsalsa208sha256_SALTBYTES]
    print("")
    print("Deriving root key with:")
    print("  opsl = ", self._ops)
    print("  meml = ", self._mem)
    print("  salt = ", hexlify(self._salt))
    self._seed = pysodium.crypto_pwhash_scryptsalsa208sha256(pysodium.crypto_sign_SEEDBYTES, password, self._salt, opslimit=self._ops, memlimit=self._mem)
    self._pk, self._sk = pysodium.crypto_sign_seed_keypair(self._seed)
    print("  Root Public Key: ", hexlify(self._pk))
    print("")



from time import time
from kafkacrypto.utils import format_exception_shim
import pysodium
import msgpack
import logging
from kafkacrypto.chain import process_chain, ProcessChainError

class Provisioners(object):
  """Class validating key requests

  Keyword Arguments:
         allowlist (array): List of allowlisted public keys (provisioners)
          denylist (array): Optional list of denylisted public keys
  """
  def __init__(self, allowlist=None, denylist=None):
    self._logger = logging.getLogger(__name__)
    self.__allowlist = allowlist
    self.__denylist = denylist

  def reencrypt_request(self, topic, cryptoexchange, msgkey=None, msgval=None):
    if (isinstance(topic,(bytes,bytearray))):
      self._logger.debug("topic provided as bytes (should be string)")
      topic = topic.decode('utf-8')
    #
    # msgval should be a msgpacked chain chaining to a provisioner.
    # The third item in the array is the public key to retransmit. The fourth item is
    # random0.
    #
    try:
      pk = None
      try:
        pk,pkprint = process_chain(msgval,topic,'key-encrypt-subscribe',allowlist=self.__allowlist,denylist=self.__denylist)
      except ProcessChainError as pce:
        raise pce
      except:
        pass
      if (pk is None):
        if not (pkprint is None):
          raise ProcessChainError("Request did not validate: ", pkprint)
        else:
          raise ValueError("Request did not validate!")
      # TODO: can add sanity checking of random0 and the public key if desired.
      msg = cryptoexchange.signed_epks(topic, epks=[pk[2]], random0=pk[3])[0]
    except Exception as e:
      self._logger.warning("".join(format_exception_shim(e)))
      return (None, None)
    return (msgkey, msg)



try:
  from kafka.future import Future as Future
except ImportError:
  import logging
  #
  # kafka-python compatible implementation of Future
  #
  class Future(object):
    done = False
    value = None
    exception = None
    exceptions_on_callbacks = False
    callbacks = []
    errorbacks = []

    def __init__(self):
      self._log = logging.getLogger(__name__)

    def retriable(self):
      try:
        return self.exception.retriable
      except:
        return False

    def success(self, val):
      assert not self.done, 'Future already completed'
      self.value = val
      self.done = True
      if self.callbacks:
        self.call_backs('callback', self.callbacks, self.value)
      return self

    def succeeded(self):
      return self.done and not bool(self.exception)

    def failure(self, e):
      assert not self.done, 'Future already completed'
      self.exception = e if type(e) is not type else e()
      assert isinstance(self.exception, BaseException), 'Future failed with no exception'
      self.done = True
      self.call_backs('errorback', self.errorbacks, self.exception)
      return self

    def failed(self):
      return self.done and bool(self.exception)

    def add_callback(self, f, *args, **kwargs):
      if args or kwargs:
        f = functools.partial(f, *args, **kwargs)
      # Execute immediately if done (without error), otherwise queue for when completed
      if self.done and not self.exception:
        self.call_backs('callback', [f], self.value)
      else:
        self.callbacks.append(f)
      return self

    def add_errback(self, f, *args, **kwargs):
      if args or kwargs:
        f = functools.partial(f, *args, **kwargs)
      # Execute immediately if done (with error), otherwise queue for later
      if self.done and self.exception:
        self.call_backs('errorback', [f], self.exception)
      else:
        self.errorbacks.append(f)
      return self

    def add_both(self, f, *args, **kwargs):
      self.add_callback(f, *args, **kwargs)
      self.add_errback(f, *args, **kwargs)
      return self

    def chain(self, future):
      self.add_callback(future.success)
      self.add_errback(future.failure)
      return self

    def call_backs(self, back_type, backs, value):
      for f in backs:
        try:
          f(value)
        except Exception as e:
          self._log.exception('Error processing %s', back_type)
          if self.exceptions_on_callbacks:
            raise e

try:
  from kafka.serializer import Deserializer as AbstractDeserializer, Serializer as AbstractSerializer
except ImportError:
  #
  # Abstract versions of SerDes
  #
  import abc

  class AbstractDeserializer(object):
    __meta__ = abc.ABCMeta

    def __init__(self, **config):
      pass

    @abc.abstractmethod
    def deserialize(self, topic, bytes_):
      pass

    def close(self):
      pass

  class AbstractSerializer(object):
    __meta__ = abc.ABCMeta

    def __init__(self, **config):
      pass

    @abc.abstractmethod
    def serialize(self, topic, value):
      pass

    def close(self):
      pass

#
# Configuration option names
# TODO: list confluent-kafka only options so they are not filtered out
#
PRODUCER_CONFIGS = [
  'bootstrap_servers',
  'client_id',
  'key_serializer',
  'value_serializer',
  'acks',
  'bootstrap_topics_filter',
  'compression_type',
  'retries',
  'batch_size',
  'linger_ms',
  'partitioner',
  'buffer_memory',
  'connections_max_idle_ms',
  'max_block_ms',
  'max_request_size',
  'metadata_max_age_ms',
  'retry_backoff_ms',
  'request_timeout_ms',
  'receive_buffer_bytes',
  'send_buffer_bytes',
  'socket_options',
  'reconnect_backoff_ms',
  'reconnect_backoff_max_ms',
  'max_in_flight_requests_per_connection',
  'security_protocol',
  'ssl_context',
  'ssl_check_hostname',
  'ssl_cafile',
  'ssl_certfile',
  'ssl_keyfile',
  'ssl_crlfile',
  'ssl_password',
  'ssl_ciphers',
  'api_version',
  'api_version_auto_timeout_ms',
  'metric_reporters',
  'metrics_num_samples',
  'metrics_sample_window_ms',
  'selector',
  'sasl_mechanism',
  'sasl_plain_username',
  'sasl_plain_password',
  'sasl_kerberos_service_name',
  'sasl_kerberos_domain_name',
  'sasl_oauth_token_provider',
]

CONSUMER_CONFIGS = [
  'bootstrap_servers',
  'client_id',
  'group_id',
  'key_deserializer',
  'value_deserializer',
  'fetch_max_wait_ms',
  'fetch_min_bytes',
  'fetch_max_bytes',
  'max_partition_fetch_bytes',
  'request_timeout_ms',
  'retry_backoff_ms',
  'reconnect_backoff_ms',
  'reconnect_backoff_max_ms',
  'max_in_flight_requests_per_connection',
  'auto_offset_reset',
  'enable_auto_commit',
  'auto_commit_interval_ms',
  'default_offset_commit_callback',
  'check_crcs',
  'metadata_max_age_ms',
  'partition_assignment_strategy',
  'max_poll_records',
  'max_poll_interval_ms',
  'session_timeout_ms',
  'heartbeat_interval_ms',
  'receive_buffer_bytes',
  'send_buffer_bytes',
  'socket_options',
  'consumer_timeout_ms',
  'security_protocol',
  'ssl_context',
  'ssl_check_hostname',
  'ssl_cafile',
  'ssl_certfile',
  'ssl_keyfile',
  'ssl_crlfile',
  'ssl_password',
  'ssl_ciphers',
  'api_version',
  'api_version_auto_timeout_ms',
  'connections_max_idle_ms',
  'metric_reporters',
  'metrics_num_samples',
  'metrics_sample_window_ms',
  'metric_group_prefix',
  'selector',
  'exclude_internal_topics',
  'sasl_mechanism',
  'sasl_plain_username',
  'sasl_plain_password',
  'sasl_kerberos_service_name',
  'sasl_kerberos_domain_name',
  'sasl_oauth_token_provider',
  'legacy_iterator',
]




from threading import Lock
from time import time
from kafkacrypto.utils import format_exception_shim, msgpack_default_pack
import pysodium
import msgpack
import logging
from kafkacrypto.chain import process_chain, ProcessChainError, key_in_list
from binascii import unhexlify
from kafkacrypto.keys import get_pks

class CryptoExchange(object):
  """Class implementing the key exchange protocol used to transmit/receive
     current data encryption keys.

  Keyword Arguments:
            chain (bytes): Certificate chain for our signing public key.
          cryptokey (obj): Object implementing the necessary public/private
                           key functions (get/sign_spk,get/use_epks).
             maxage (int): Maximum validity time for sent messages in sec.
                           Optional, default: 86400 (= 1 day)
        randombytes (int): Number of random bytes to add in constructing
                           shared secrets. Optional, minimum default: 32
        allowlist (array): Array of allowlisted public keys. Optional,
                           default: None
         denylist (array): Array of denylisted public keys. Optional,
                           default: None
  """
  #
  # Per instance, defined in init
  #      __maxage: Maximum age (seconds)
  # __randombytes: Number of random bytes added to exchange
  #     __randoms: dictionary of arrays of random values used during
  #                key exchange queries, indexed by topic
  #   __spk_chain: signing public key chain to root of trust, as array
  #   __cryptokey: Provider of operations involving private keys
  #
  def __init__(self, chain, cryptokey, maxage=0, randombytes=0, allowlist=None, denylist=None):
    self._logger = logging.getLogger(__name__)
    self.__maxage = maxage if (maxage!=None and maxage>0) else 86400
    self.__randombytes = randombytes if randombytes>=32 else 32
    self.__randoms = {}
    self.__randoms_lock = Lock()
    self.__cryptokey = cryptokey
    self.__spk_direct_request = True
    self.__spk_chain = []
    self.__spk_chain_lock = Lock()
    self.__allowdenylist_lock = Lock()
    if not (allowlist is None):
      self.__allowlist = allowlist
    else:
      self.__allowlist = []
    if not (denylist is None):
      self.__denylist = denylist
    else:
      self.__denylist = []
    self.__update_spk_chain(chain)

  def encrypt_keys(self, keyidxs, keys, topic, msgval=None):
    if (isinstance(topic,(bytes,bytearray))):
      self._logger.debug("passed a topic in bytes (should be string)")
      topic = topic.decode('utf-8')
    #
    # msgval should be a msgpacked chain.
    # The public key in the array is the public key to send the key to, using a
    # common DH-derived key between that public key and our private encryption key.
    # Then there is at least one more additional item, random bytes:
    # (3) random bytes
    # Currently items after this are ignored, and reserved for future use.
    #
    try:
      with self.__allowdenylist_lock:
        pk,_ = process_chain(msgval,topic,'key-encrypt-request',allowlist=self.__allowlist,denylist=self.__denylist)
      # Construct shared secret as sha256(topic || random0 || random1 || our_private*their_public)
      if not isinstance(pk[2], (list,)):
        # legacy format for key requests
        pks = [pk[2]]
      else:
        pks = pk[2]
      # (re)generate and then immediately use epks
      self.__cryptokey.get_epks(topic,'encrypt_keys')
      eks,epk = self.__cryptokey.use_epks(topic,'encrypt_keys',pks)
      if len(eks) < 1:
        self._logger.info("No compatible key exchange versions found.")
        return None # No compatible versions
      ek = eks[0]
      random0 = pk[3]
      random1 = pysodium.randombytes(self.__randombytes)
      ss = pysodium.crypto_hash_sha256(topic.encode('utf-8') + random0 + random1 + ek)[0:pysodium.crypto_secretbox_KEYBYTES]
      nonce = pysodium.randombytes(pysodium.crypto_secretbox_NONCEBYTES)
      # encrypt keys and key indexes (MAC appended, nonce prepended)
      msg = []
      for i in range(0,len(keyidxs)):
        msg.append(keyidxs[i])
        msg.append(keys[i])
      msg = msgpack.packb(msg, default=msgpack_default_pack, use_bin_type=True)
      msg = nonce + pysodium.crypto_secretbox(msg,nonce,ss)
      # this is then put in a msgpack array with the appropriate max_age, poison, and public key(s)
      poison = msgpack.packb([['topics',[topic]],['usages',['key-encrypt']]], default=msgpack_default_pack, use_bin_type=True)
      msg = msgpack.packb([time()+self.__maxage,poison,[epk[0]],[random0,random1],msg], default=msgpack_default_pack, use_bin_type=True)
      # and signed with our signing key
      msg = self.__cryptokey.sign_spk(msg)
      # and finally put as last member of a msgpacked array chaining to ROT
      with self.__spk_chain_lock:
        tchain = self.__spk_chain.copy()
        if (len(tchain) == 0):
          poison = msgpack.packb([['topics',[topic]],['usages',['key-encrypt']],['pathlen',1]], default=msgpack_default_pack, use_bin_type=True)
          lastcert = msgpack.packb([time()+self.__maxage,poison,self.__cryptokey.get_spk()], default=msgpack_default_pack, use_bin_type=True)
          _,tempsk = pysodium.crypto_sign_seed_keypair(unhexlify(b'4c194f7de97c67626cc43fbdaf93dffbc4735352b37370072697d44254e1bc6c'))
          tchain.append(pysodium.crypto_sign(lastcert,tempsk))
          provision = msgpack.packb([msgpack.packb([0,b'\x90',self.__cryptokey.get_spk()], default=msgpack_default_pack),self.__cryptokey.sign_spk(lastcert)], default=msgpack_default_pack, use_bin_type=True)
          self._logger.warning("Current signing chain is empty. Use %s to provision access and then remove temporary root of trust from allowedlist.", provision.hex())
      tchain.append(msg)
      msg = msgpack.packb(tchain, default=msgpack_default_pack, use_bin_type=True)
    except Exception as e:
      self._logger.warning("".join(format_exception_shim(e)))
      return None
    return msg

  def decrypt_keys(self, topic, msgval=None):
    if (isinstance(topic,(bytes,bytearray))):
      self._logger.debug("passed a topic in bytes (should be string)")
      topic = topic.decode('utf-8')
    #
    # msgval should be a msgpacked chain.
    # The public key in the array is a set of public key(s) to combine with our
    # encryption key to get the secret to decrypt the key. If we do not
    # have an encryption key for the topic, we cannot do it until we have one.
    # The next item is then the pair of random values for generating the shared
    # secret, followed by the actual key message.
    #
    try:
      with self.__allowdenylist_lock:
        pk,pkprint = process_chain(msgval,topic,'key-encrypt',allowlist=self.__allowlist,denylist=self.__denylist)
      if (len(pk) < 5):
        if not (pkprint is None):
          raise ProcessChainError("Unexpected number of chain elements:", pkprint)
        else:
          raise ValueError("Unexpected number of chain elements!")
      random0 = pk[3][0]
      with self.__randoms_lock:
        if not (topic in self.__randoms) or not (random0 in self.__randoms[topic]):
          self._logger.info("unknown (or already used) random0 value in decrypt_keys: %s vs %s", str(random0), str(self.__randoms[topic]) if topic in self.__randoms else "")
          return None
      random1 = pk[3][1]
      nonce = pk[4][0:pysodium.crypto_secretbox_NONCEBYTES]
      msg = pk[4][pysodium.crypto_secretbox_NONCEBYTES:]
      eks,_ = self.__cryptokey.use_epks(topic, 'decrypt_keys', pk[2], clear=False)
      for ck in eks:
        # Construct candidate shared secrets as sha256(topic || random0 || random1 || our_private*their_public)
        ss = pysodium.crypto_hash_sha256(topic.encode('utf-8') + random0 + random1 + ck)[0:pysodium.crypto_secretbox_KEYBYTES]
        # decrypt and return key
        try:
          msg = msgpack.unpackb(pysodium.crypto_secretbox_open(msg,nonce,ss),raw=True)
          rvs = {}
          for i in range(0,len(msg),2):
            rvs[msg[i]] = msg[i+1]
          if len(rvs) < 1 or 2*len(rvs) != len(msg):
            raise ValueError
        except:
          pass
        else:
          # clear the esk/epk we just used
          self.__cryptokey.use_epks(topic, 'decrypt_keys', [])
          with self.__randoms_lock:
            self.__randoms[topic].remove(random0)
          return rvs
      self._logger.info("no valid decryption keys computed in decrypt_keys, %s tried", str(len(eks)))
    except Exception as e:
      self._logger.warning("".join(format_exception_shim(e)))
      pass
    return None

  def signed_epks(self, topic, epks=None, random0=None):
    if (isinstance(topic,(bytes,bytearray))):
      self._logger.debug("passed a topic in bytes (should be string)")
      topic = topic.decode('utf-8')
    #
    # returns the public key of a current or given ephemeral key for the specified topic
    # (generating a new one if not present), signed by our signing key,
    # with a fresh random value, and with the chain to the ROT prepended.
    #
    try:
      if epks is None:
        epks = self.__cryptokey.get_epks(topic,'decrypt_keys')
      if random0 is None:
        random0 = pysodium.randombytes(self.__randombytes)
        with self.__randoms_lock:
          if not (topic in self.__randoms):
            self.__randoms[topic] = []
          self.__randoms[topic].append(random0) # store to check later
      # we allow either direct-to-producer or via-controller key establishment
      poison = msgpack.packb([['topics',[topic]],['usages',['key-encrypt-request','key-encrypt-subscribe']]], default=msgpack_default_pack, use_bin_type=True)
      rv = []
      for epk in epks:
        # create message for each independent set of epks
        msg = msgpack.packb([time()+self.__maxage,poison,epk,random0], default=msgpack_default_pack, use_bin_type=True)
        # and signed with our signing key
        msg = self.__cryptokey.sign_spk(msg)
        # and finally put as last member of a msgpacked array chaining to ROT
        with self.__spk_chain_lock:
          tchain = self.__spk_chain.copy()
          if (len(tchain) == 0):
            # Use default for direct use when empty.
            self.__spk_direct_request = True
            poison = msgpack.packb([['topics',[topic]],['usages',['key-encrypt-request','key-encrypt-subscribe']],['pathlen',1]], default=msgpack_default_pack, use_bin_type=True)
            lastcert = msgpack.packb([time()+self.__maxage,poison,self.__cryptokey.get_spk()], default=msgpack_default_pack, use_bin_type=True)
            _,tempsk = pysodium.crypto_sign_seed_keypair(unhexlify(b'4c194f7de97c67626cc43fbdaf93dffbc4735352b37370072697d44254e1bc6c'))
            tchain.append(pysodium.crypto_sign(lastcert,tempsk))
            provision = msgpack.packb([msgpack.packb([0,b'\x90',self.__cryptokey.get_spk()]),self.__cryptokey.sign_spk(lastcert)], default=msgpack_default_pack, use_bin_type=True)
            self._logger.warning("Current signing chain is empty. Use %s to provision access and then remove temporary root of trust from allowedlist.", provision.hex())
        tchain.append(msg)
        msg = msgpack.packb(tchain, default=msgpack_default_pack, use_bin_type=True)
        rv.append(msg)
      return rv
    except Exception as e:
      self._logger.warning("".join(format_exception_shim(e)))
      pass
    return None

  def add_allowlist(self, allow):
    self.__allowdenylist_lock.acquire()
    try:
      pk,_ = process_chain(allow,None,'key-allowlist',allowlist=self.__allowlist,denylist=self.__denylist)
      if (len(pk) >= 4):
        apk = msgpack.unpackb(pk[3],raw=True)
        apk[2] = get_pks(apk[2])
        if pk[2] != apk[2]:
          self._logger.info("Mismatch in keys for allowlist %s vs %s.",str(pk[2]),str(apk[2]))
          raise ValueError("Mismatch in keys for allowlist.")
        if key_in_list(apk[2],self.__allowlist)==None:
          self.__allowlist.append(pk[3])
        else:
          self._logger.info("Key %s already in allowlist",str(apk[2]))
          raise ValueError("Key already in allowlist!")
        self._logger.warning("Added key %s to allowlist",str(apk[2]))
        return pk[3]
    except Exception as e:
      self._logger.warning("".join(format_exception_shim(e)))
      return None
    finally:
      self.__allowdenylist_lock.release()

  def add_denylist(self, deny):
    self.__allowdenylist_lock.acquire()
    try:
      pk,_ = process_chain(deny,None,'key-denylist',allowlist=self.__allowlist,denylist=self.__denylist)
      if (len(pk) >= 4):
        apk = msgpack.unpackb(pk[3],raw=True)
        apk[2] = get_pks(apk[2])
        if pk[2] != apk[2]:
          self._logger.info("Mismatch in keys for denylist %s vs %s.",str(pk[2]),str(apk[2]))
          raise ValueError("Mismatch in keys for denylist.")
        if key_in_list(apk[2],self.__denylist)==None:
          self.__denylist.append(pk[3])
        else:
          self._logger.info("Key %s already in denylist",str(pk[2]))
          raise ValueError("Key already in denylist!")
        self._logger.warning("Added key %s to denylist",str(pk[2]))
       	return pk[3]
    except Exception as e:
      self._logger.warning("".join(format_exception_shim(e)))
      return None
    finally:
      self.__allowdenylist_lock.release()

  def direct_request_spk_chain(self):
    with self.__spk_chain_lock:
      rv = self.__spk_direct_request
    return rv

  def replace_spk_chain(self, newchain):
    # update_spk_chain captures any exceptions and returns None when they happen.
    newchain = self.__update_spk_chain(newchain)
    return newchain

  def __update_spk_chain(self, newchain):
    #
    # We have a new candidate chain to replace the current one (if any). We
    # first must check it is a valid chain ending in our signing public key.
    #
    if (newchain is None):
      return None
    try:
      with self.__allowdenylist_lock:
        pk,pkprint = process_chain(newchain,None,None,allowlist=self.__allowlist,denylist=self.__denylist)
      if (len(pk) < 3 or pk[2] != self.__cryptokey.get_spk()):
        if not (pkprint is None):
          raise ProcessChainError("New chain does not match current signing publickey:", pkprint)
        else:
          raise ValueError("New chain does not match current signing public key,")
      # If we get here, it is a valid chain. So now we need
      # to see if it is "superior" than our current chain.
      # This means a larger minimum max_age.
      min_max_age = 0
      with self.__spk_chain_lock:
        for cpk in self.__spk_chain:
          if (min_max_age == 0 or cpk[0]<min_max_age):
            min_max_age = cpk[0]
      if (pk[0] < min_max_age):
        raise ProcessChainError("New chain has shorter expiry time than current chain.", pkprint)
      with self.__spk_chain_lock:
        self.__spk_chain = msgpack.unpackb(newchain,raw=True)
        self._logger.warning("Utilizing new chain: %s", str(pkprint))
      # Check if we are capable of direct key requests (default is "no")
      with self.__spk_chain_lock:
        self.__spk_direct_request = False
        self._logger.debug("Defaulting new chain to not handle direct requests.")
      try:
        with self.__allowdenylist_lock:
          pk,_ = process_chain(newchain,None,'key-encrypt-request',allowlist=self.__allowlist,denylist=self.__denylist)
        if len(pk) >= 3:
          with self.__spk_chain_lock:
            self._logger.info("  New chain supports direct key requests. Enabling.")
            self.__spk_direct_request = True
      except Exception as e:
        # exceptions when checking direct mean it is not supported
        with self.__spk_chain_lock:
          self._logger.info("  New chain does not support direct key requests. Disabling.")
          self.__spk_direct_request = False
      return newchain
    except Exception as e:
      self._logger.warning("".join(format_exception_shim(e)))
      return None



import msgpack
import pysodium
try:
  # until more mature implementations are available, use Open Quantum Safe
  import oqs
except ImportError:
  pass

class PublicKeyError(Exception):
  """
  Error converting to a public key
  """

class SecretKeyError(Exception):
  """
  Error converting to a secret key
  """

def get_pks(pk0):
  if isinstance(pk0, (SignPublicKey,KEMPublicKey)):
    return pk0
  elif isinstance(pk0, (list,tuple)):
    if isinstance(pk0[0], (int,)): # versioned
      try:
        return SignPublicKey(pk0)
      except PublicKeyError:
        return KEMPublicKey(pk0)
    else:
      # list of public keys
      rv = []
      for pk in pk0:
        rv.append(get_pks(pk))
      return rv
  else:
    # implicit version 1, treat as sign public key for now
    return SignPublicKey([1, pk0])

# Public keys used for signing.
# version is an integer that determines the meaning of keys.
#  --> except version 1, values should be unique between Sign and KEM classes
# version   keys
# -------   -------
#    1      bytes representing an Ed25519 public key
class SignPublicKey(object):
  version = 0
  keys = b''
  def __init__(self, pk0):
    if isinstance(pk0, (SignPublicKey,)):
      self.version = pk0.version
      self.keys = pk0.keys
    elif isinstance(pk0, (list,tuple)):
      if isinstance(pk0[0], (int,)) and pk0[0] == 1: # currently only version 1 supported
        self.version = pk0[0]
        self.keys = pk0[1]
      else:
        # list of public keys
        raise PublicKeyError()
    else:
      # implicit version 1
      self.version = 1
      self.keys = pk0
    # for ctypes calling compatibility (see also __len__ and __getitem__ below)
    self._as_parameter_ = bytes(self)
  def __bytes__(self):
    if self.version == 1:
      return self.keys
    else:
      return msgpack.packb([self.version, self.keys],use_bin_type=True)
  def __str__(self):
    if self.version == 1:
      return "(Ed25519, " + self.keys.hex() + ")"
    else:
      return "(Unparsable)"
  def __eq__(self, pk2):
    if isinstance(pk2, (SignPublicKey,)) and self.version == pk2.version:
      if self.version == 1 and self.keys == pk2.keys:
        return True
    return False
  def __len__(self):
    return len(self._as_parameter_)
  def __getitem__(self, index):
    return self._as_parameter_[index]
  def __iter__(self):
    return iter([self.version, self.keys])
  def msgpack_pack(self):
    if self.version == 1:
      return self.keys
    else:
      return list(self)
  def crypto_sign_open(self, cso):
    if self.version == 1:
      return pysodium.crypto_sign_open(cso,self.keys)
    raise PublicKeyError

# Public keys used for key exchange.
# version is an integer that determines the meaning of keys.
#  --> except version 1, values should be unique between Sign and KEM classes
# version   keys
# -------   -------
#    1      bytes representing an Curve25519 public key
#    2      list of two items: Curve25519 pk bytes, sntrup761 pk bytes
#    3      list of two items: Curve25519 pk bytes, sntrup761 ct bytes
class KEMPublicKey(object):
  version = 0
  keys = b''
  def __init__(self, pk0):
    if isinstance(pk0, (SignPublicKey,)) and pk0.version == 1: # we allow conversion of implicit version 1 keys between Sign and KEM types
      self.version = pk0.version
      self.keys = pk0.keys
    elif isinstance(pk0, (KEMPublicKey,)):
      self.version = pk0.version
      self.keys = pk0.keys
    elif isinstance(pk0, (KEMSecretKey,)): # generating public key from secret key is allowed
      if pk0.version == 3:
        self.version = 3
        # compute Curve25519 public key in situ, include sntrup761 encapped ciphertext
        self.keys = [pysodium.crypto_scalarmult_curve25519_base(pk0.keys[0]),pk0.keys[1][2]]
      elif pk0.version == 2:
        self.version = 2
        # compute Curve25519 public key in situ, include sntrup761 public key
        self.keys = [pysodium.crypto_scalarmult_curve25519_base(pk0.keys[0]),pk0.keys[1][1]]
      else:
        # default version 1
        self.version = 1
        self.keys = pysodium.crypto_scalarmult_curve25519_base(pk0.keys)
    elif isinstance(pk0, (list,tuple)):
      if isinstance(pk0[0], (int,)) and (pk0[0] == 1 or pk0[0] == 2 or pk0[0] == 3): # Versions 1-3 only
        self.version = pk0[0]
        self.keys = pk0[1]
      else:
        # list of public keys
        raise PublicKeyError()
    else:
      # implicit version 1
      self.version = 1
      self.keys = pk0
    # for ctypes calling compatibility (see also __len__ and __getitem__ below)
    self._as_parameter_ = bytes(self)
  def __bytes__(self):
    if self.version == 1:
      return self.keys
    else:
      return msgpack.packb([self.version, self.keys],use_bin_type=True)
  def __str__(self):
    if self.version == 1:
      return "(Curve25519-pk, " + self.keys.hex() + ")"
    elif self.version == 2:
      return "(Curve25519-sntrup761-pk, [" + self.keys[0].hex() + "," + self.keys[1].hex() + "])"
    elif self.version == 3:
      return "(Curve25519-sntrup761-ct, [" + self.keys[0].hex() + "," + self.keys[1].hex() + "])"
    else:
      return "(Unparsable)"
  def __eq__(self, pk2):
    if isinstance(pk2, (KEMPublicKey,)) and self.version == pk2.version:
      if self.version == 1 and self.keys == pk2.keys:
        return True
      elif (self.version == 2 or self.version == 3) and self.keys[0] == pk2.keys[0] and self.keys[1] == pk2.keys[1]:
        return True
    return False
  def __len__(self):
    return len(self._as_parameter_)
  def __getitem__(self, index):
    return self._as_parameter_[index]
  def __iter__(self):
    return iter([self.version, self.keys])
  def get_esk_version(self):
    if self.version == 3:
      return 2
    else:
      return self.version
  def msgpack_pack(self):
    if self.version == 1:
      return self.keys
    else:
      return list(self)

# Secret keys used for key exchange.
# version is an integer that determines the meaning of keys.
#  --> should have the same versions as KEMPublicKey
# version   keys
# -------   -------
#    1      bytes representing an Curve25519 secret key
#   2,3     list of two items: Curve25519 bytes, sntrup761 bytes list [sk,pk,ct (v. 3 only)]
class KEMSecretKey(object):
  version = 0
  keys = b''
  def __init__(self, pk0):
    if isinstance(pk0, (KEMSecretKey,)):
      self.version = pk0.version
      self.keys = pk0.keys
    elif isinstance(pk0, (list,tuple)):
      if isinstance(pk0[0], (int,)) and (pk0[0] == 1 or pk0[0] == 2 or pk0[0] == 3): # versions 1-3 only
        self.version = pk0[0]
        self.keys = pk0[1]
        if self.version == 2 or self.version == 3:
          self.sntrup761 = oqs.KeyEncapsulation("sntrup761", secret_key=self.keys[1][0])
      else:
        # list of secret keys
        raise SecretKeyError()
    elif isinstance(pk0,(bytes,bytearray)) and len(pk0) == 32:
      # implicit version 1
      self.version = 1
      self.keys = pk0
    else:
      # assume version is what was passed
      if pk0 == 3:
        raise SecretKeyError() # cannot create type-3 directly, requires a complete_kem call
      elif pk0 == 2:
        self.version = 2
        self.keys = [pysodium.randombytes(pysodium.crypto_scalarmult_curve25519_BYTES)] # Curve25519 key first
        self.sntrup761 = oqs.KeyEncapsulation("sntrup761")
        pk = self.sntrup761.generate_keypair()
        self.keys += [[self.sntrup761.export_secret_key(),pk]] # sntrup761 key second (store both secret and public key)
      else:
        # default to version 1
        self.version = 1
        self.keys = pysodium.randombytes(pysodium.crypto_scalarmult_curve25519_BYTES)
    # for ctypes calling compatibility (see also __len__ and __getitem__ below)
    self._as_parameter_ = bytes(self)
  def __bytes__(self):
    if self.version == 1:
      return self.keys
    else:
      return msgpack.packb([self.version, self.keys],use_bin_type=True)
  def __str__(self):
    if self.version == 1:
      return "(Curve25519-Secret, " + self.keys.hex() + ")"
    elif self.version == 2:
      return "(Curve25519-sntrup761-Secret, [" + self.keys[0].hex() + ",[" + self.keys[1][0].hex() + "," + self.keys[1][1].hex() + "]])"
    elif self.version == 3:
      return "(Curve25519-sntrup761-Secret-ct, [" + self.keys[0].hex() + ",[" + self.keys[1][0].hex() + "," + self.keys[1][1].hex() + "," + self.keys[1][2].hex() + "]])"
    else:
      return "(Unparsable)"
  def __eq__(self, pk2):
    if isinstance(pk2, (KEMSecretKey,)) and self.version == pk2.version:
      if self.version == 1 and self.keys == pk2.keys:
        return True
      elif self.version == 2 and self.keys[0] == pk2.keys[0] and self.keys[1][0] == pk2.keys[1][0] and self.keys[1][1] == pk2.keys[1][1]:
        return True
      elif self.version == 3 and self.keys[0] == pk2.keys[0] and self.keys[1][0] == pk2.keys[1][0] and self.keys[1][1] == pk2.keys[1][1] and self.keys[1][2] == pk2.keys[1][2]:
        return True
    return False
  def __len__(self):
    return len(self._as_parameter_)
  def __getitem__(self, index):
    return self._as_parameter_[index]
  def __iter__(self):
    return iter([self.version, self.keys])
  def msgpack_pack(self):
    if self.version == 1:
      return self.keys
    else:
      return list(self)
  def complete_kem(self, pubkey):
    if self.version == 1 and pubkey.version == 1:
      return pysodium.crypto_scalarmult_curve25519(self.keys,pubkey.keys)
    elif (self.version == 2 or self.version == 3) and pubkey.version == 2:
      part0 = pysodium.crypto_scalarmult_curve25519(self.keys[0],pubkey.keys[0])
      ct, part1 = self.sntrup761.encap_secret(pubkey.keys[1])
      if self.version == 2:
        self.version = 3
        self.keys[1].append(ct)
      else: # self.version is already 3
        self.keys[1][2] = ct
      return part0 + part1
    elif self.version == 2 and pubkey.version == 3: # cannot do 3 with 3 since we don't have partner pk
      part0 = pysodium.crypto_scalarmult_curve25519(self.keys[0],pubkey.keys[0])
      part1 = self.sntrup761.decap_secret(pubkey.keys[1])
      return part0 + part1
    else:
      raise SecretKeyError()



from threading import Lock
import pysodium
import logging
import msgpack
from os import path
from kafkacrypto.utils import msgpack_default_pack
from kafkacrypto.keys import SignPublicKey, KEMPublicKey, KEMSecretKey

class CryptoKey(object):
  """Class utilizing file-backed storage to store (long-term) private key
     material for identity, and memory-backed storage to store (short-term)
     ephemeral private key material.

  Keyword Arguments:
          file (str,file): Filename or File IO object for storing crypto info.
                           Must be readable once.
  """
  #
  # Per instance, defined in init
  #      __file: File object
  #       __spk: signing public key
  #       __ssk: signing private (secret) key
  #        __ek: private key for (un)wrapping opaque bytes
  #  __ephk_legacy: if true, and if version 1 included in __ephk_ver, then send
  #                 new key requests in legacy format (replies to legacy format
  #                 requests are not affected by this setting)
  #  __ephk_ver: ephemeral key exchange key type (KEMPublic/SecretKey version(s))
  # Generated ephemerially, on demand:
  #       __esk: dict of encrypting private (secret) keys (by topic and usage and version)
  #
  def __init__(self, file):
    self._logger = logging.getLogger(__name__)
    if (isinstance(file, (str))):
      if (not path.exists(file)):
        self.__init_cryptokey(file)
      with open(file, 'rb') as f:
        data = f.read()
    else:
      data = file.read()
    datalen = len(data)
    # this should be legacy
    contents = None
    while len(data) and contents is None:
      try:
        contents = msgpack.unpackb(data,raw=True)
      except msgpack.exceptions.ExtraData:
        data = data[:-1]
    if len(data) != datalen:
      self._logger.warning("Cryptokey file had extraneous bytes at end, attempting load anyways.")
    if len(contents) == 2:
      # unversioned legacy format, so update
      self._logger.warning("Cryptokey file is unversioned.")
      contents = [1] + contents + [True,[1]]
      if (isinstance(file, (str))):
        self._logger.warning("Cryptokey file updating from unversioned to versioned format.")
        with open(file, 'wb') as f:
          f.write(msgpack.packb(contents, default=msgpack_default_pack, use_bin_type=True))
    self.__eklock = Lock()
    self.__esk = {}
    if contents[0] == 1:
      # version 1
      self.__ssk = contents[1]
      self.__spk = SignPublicKey(pysodium.crypto_sign_sk_to_pk(self.__ssk))
      self.__ek = contents[2]
      self.__ephk_legacy = contents[3]
      self.__ephk_ver = contents[4]

  def get_spk(self):
    return self.__spk

  def sign_spk(self, msg):
    if self.__spk.version == 1:
      return pysodium.crypto_sign(msg, self.__ssk)
    return None

  def get_epks(self, topic, usage):
    #
    # returns the public key(s) of a new ephemeral encryption key for the specified topic
    # each returned value of the list is to be sent in a single message
    #
    if (isinstance(topic,(bytes,bytearray))):
      self._logger.debug("passed a topic in bytes (should be string)")
      topic = topic.decode('utf-8')
    if (isinstance(usage,(bytes,bytearray))):
      self._logger.debug("passed a usage in bytes (should be string)")
      usage = usage.decode('utf-8')
    with self.__eklock:
      self.__generate_esk(topic, usage)
      rv0 = []
      rv = []
      for v in self.__esk[topic][usage]:
        rv0.append(KEMPublicKey(self.__esk[topic][usage][v]))
      rv.append(rv0)
      if self.__ephk_legacy and 1 in self.__esk[topic][usage]:
        rv.append(KEMPublicKey(self.__esk[topic][usage][1]))
      return rv

  def use_epks(self, topic, usage, pks, clear=True):
    rv = []
    rvp = []
    if (isinstance(topic,(bytes,bytearray))):
      self._logger.debug("passed a topic in bytes (should be string)")
      topic = topic.decode('utf-8')
    if (isinstance(usage,(bytes,bytearray))):
      self._logger.debug("passed a usage in bytes (should be string)")
      usage = usage.decode('utf-8')
    with self.__eklock:
      if not topic in self.__esk or not usage in self.__esk[topic]:
        return (rv,rvp)
      for pk in pks:
        kpk = KEMPublicKey(pk)
        if kpk.get_esk_version() in self.__esk[topic][usage]:
          rv.append(self.__esk[topic][usage][kpk.get_esk_version()].complete_kem(kpk))
          rvp.append(KEMPublicKey(self.__esk[topic][usage][kpk.get_esk_version()]))
      if clear:
        self.__remove_esk(topic, usage)
    return (rv,rvp)

  def wrap_opaque(self, crypto_opaque):
    nonce = pysodium.randombytes(pysodium.crypto_secretbox_NONCEBYTES)
    return nonce+pysodium.crypto_secretbox(crypto_opaque,nonce,self.__ek)

  def unwrap_opaque(self, crypto_opaque):
    try:
      return pysodium.crypto_secretbox_open(crypto_opaque[pysodium.crypto_secretbox_NONCEBYTES:],crypto_opaque[0:pysodium.crypto_secretbox_NONCEBYTES], self.__ek)
    except:
      return None

  def __generate_esk(self, topic, usage):
    # ephemeral keys are use once only, so always ok to overwrite
    # caller must hold self.__eklock prior to calling
    if not topic in self.__esk:
      self.__esk[topic] = {}
    if not usage in self.__esk[topic]:
      self.__esk[topic][usage] = {}
    for v in self.__ephk_ver:
      self.__esk[topic][usage][v] = KEMSecretKey(v)

  def __remove_esk(self, topic, usage):
    # caller must hold self.__eklock prior to calling
    self.__esk[topic].pop(usage)

  def __init_cryptokey(self, file):
    self._logger.warning("Initializing new CryptoKey file %s", file)
    pk,sk = pysodium.crypto_sign_keypair()
    self._logger.warning("  Public key: %s", pysodium.crypto_sign_sk_to_pk(sk).hex())
    with open(file, "wb") as f:
      f.write(msgpack.packb([1,sk,pysodium.randombytes(pysodium.crypto_secretbox_KEYBYTES),True,[1]], default=msgpack_default_pack, use_bin_type=True))
    self._logger.warning("  CryptoKey Initialized. Provisioning required for successful operation.")



import copy
import inspect
import logging
from time import time
from threading import Lock
from confluent_kafka import Producer, Consumer, TopicPartition as TopicPartitionOffset, OFFSET_BEGINNING, OFFSET_END, TIMESTAMP_NOT_AVAILABLE, KafkaException, KafkaError
from kafkacrypto.common_wrapper import Future
from kafkacrypto.exceptions import KafkaCryptoWrapperError
from collections import namedtuple

TopicPartition = namedtuple("TopicPartition", ["topic", "partition"])
OffsetAndMetadata = namedtuple("OffsetAndMetadata", ["offset", "metadata"])
Message = namedtuple("Message",
  ["topic", "partition", "offset", "timestamp", "headers", "key", "value"])
RecordMetadata = namedtuple('RecordMetadata',
  ['topic', 'partition', 'topic_partition', 'offset', 'timestamp', 'checksum', 'serialized_key_size', 'serialized_value_size', 'serialized_header_size'])

class FutureTimeoutError(Exception):
  """
  Timeout waiting for future to complete
  """

class FutureRecordMetadata(Future):
  """
  Our implementation of FutureRecordMetadata so producer callbacks can be used
  """
  def __init__(self, producer, value_len, key_len):
    super().__init__()
    self._producer = producer
    self.key_len = key_len
    self.value_len = value_len

  def base_callback(self, err, msg):
    self._producer._log.debug("Entering base callback with err=%s and msg=%s.", str(err), str(msg))
    if err is not None:
      self._producer._log.debug("Callback failed (non-None err).")
      super().failure(KafkaException(err))
    elif msg is None:
      self._producer._log.debug("Callback failed (None msg).")
      super().failure(KafkaException(KafkaError.UNKNOWN,'null msg'))
    elif msg.error() is not None:
      self._producer._log.debug("Callback failed (non-None msg.error).")
      super().failure(KafkaException(msg.error()))
    else:
      # success
      self._producer._log.debug("Callback success.")
      metadata = RecordMetadata(msg.topic(), msg.partition(), TopicPartition(msg.topic(), msg.partition()),
                                msg.offset(), msg.timestamp()[1] if msg.timestamp()[0]!=TIMESTAMP_NOT_AVAILABLE else int(time()*1000),
                                None, self.key_len, self.value_len, -1)
      super().success(metadata)
    self._producer._log.debug("Finished with base callback with err=%s and msg=%s.", str(err), str(msg))

  def get(self, timeout="default", timeout_jiffy=0.1):
    # timeout = None is infinite timeout, for compatibility with kafka-python's FutureRecordMetadata
    # timeout = "default" (or any non-number) should never be passed by callers, but is used here
    #           to indicate use of configured default.
    if timeout is None:
      last_time = 9223372036854775807
      timeout = last_time-time()
    elif timeout == "default":
      timeout = self._producer.config['produce_timeout']
      last_time = time()+timeout
    else:
      last_time = time()+timeout
    self._producer._log.debug("Entering FutureRecordMetadata get with timeout=%s.", str(timeout))
    orig_timeout = timeout
    while not self.is_done and timeout>0:
      self._producer.poll(min(timeout,timeout_jiffy))
      timeout = last_time-time()
      self._producer._log.debug("FutureRecordMetadata polling done, remaining timeout=%s.", str(timeout))
    if not self.is_done:
      raise FutureTimeoutError("Timeout after waiting for %s secs." % (orig_timeout,))
    if super().failed():
      raise self.exception
    self._producer._log.debug("Finished with FutureRecordMetadata get")
    return self.value


class KafkaConsumer(Consumer):
  """
  Thin wrapper around confluent-kafka providing methods unique to kafka-python
  to allow them to be used interchangeably. Unless otherwise noted, parameters should
  be supplied as expected by the underlying confluent_kafka classes. Use at your own risk!
  """
  CONFIG_MAP = { 'ssl_cafile': 'ssl.ca.location',
                 'ssl_certfile': 'ssl.certificate.location',
                 'ssl_keyfile': 'ssl.key.location',
                 'ssl_crlfile': 'ssl.crl.location',
                 'ssl_password': 'ssl.key.password',
                 'ssl_ciphers': 'ssl.cipher.suites',
                 'max_partition_fetch_bytes': 'message.max.bytes', # this maps the hard limit functionality appropriately
                 'bootstrap_servers': 'bootstrap.servers',
                 'client_id': 'client.id',
                 'group_id': 'group.id',
                 'fetch_min_bytes': 'fetch.min.bytes',
                 'fetch_max_wait_ms': 'fetch.wait.max.ms',
                 'fetch_max_bytes': 'fetch.max.bytes',
                 'request_timeout_ms': 'request.timeout.ms',
                 'retry_backoff_ms': 'retry.backoff.ms',
                 'reconnect_backoff_ms': 'reconnect.backoff.ms',
                 'reconnect_backoff_max_ms': 'reconnect.backoff.max.ms',
                 'max_in_flight_requests_per_connection': 'max.in.flight.requests.per.connection',
                 'auto_offset_reset': 'auto.offset.reset',
                 'enable_auto_commit': 'enable.auto.commit',
                 'auto_commit_interval_ms': 'auto.commit.interval.ms',
                 'check_crcs': 'check.crcs',
                 'metadata_max_age_ms': 'metadata.max.age.ms',
                 'partition_assignment_strategy': 'partition.assignment.strategy',
                 'max_poll_interval_ms': 'max.poll.interval.ms',
                 'session_timeout_ms': 'session.timeout.ms',
                 'heartbeat_interval_ms': 'heartbeat.interval.ms',
                 'receive_buffer_bytes': 'socket.receive.buffer.bytes',
                 'send_buffer_bytes': 'socket.send.buffer.bytes',
                 'security_protocol': 'security.protocol',
                 'sasl_mechanism': 'sasl.mechanism',
                 'sasl_plain_username': 'sasl.username',
                 'sasl_plain_password': 'sasl.password',
                 'sasl_kerberos_service_name': 'sasl.kerberos.service.name',
               }
  CONFIG_MAP_LOCAL = { 'max_poll_records': 500,
                     }
  CONFIG_MAP_NULL = [ 'ssl_context',
                      'socket_options',
                      'default_offset_commit_callback',
                      'consumer_timeout_ms',
                      'ssl_check_hostname',
                      'api_version',
                      'api_version_auto_timeout_ms',
                      'connections_max_idle_ms',
                      'metrics_reporters',
                      'metrics_num_samples',
                      'metrics_sample_window_ms',
                      'selector',
                      'exclude_internal_topics',
                      'sasl_kerberos_domain_name',
                      'sasl_oauth_token_provider',
                    ]
  def __init__(self, *topics, **configs):
    self._log = logging.getLogger(__name__)
    self.raw_config = configs
    self.cf_config = {}
    self.config = {}
    self.kds = lambda topic, _bytes: _bytes
    self.vds = lambda topic, _bytes: _bytes
    for k in self.CONFIG_MAP_LOCAL.keys():
      self.config[k] = self.CONFIG_MAP_LOCAL[k]
    for k in self.raw_config.keys():
      if (k == 'key_deserializer'):
        if not (self.raw_config[k] is None) and (not hasattr(self.raw_config[k], 'deserialize') or not inspect.isroutine(self.raw_config[k].deserialize)):
          self.kds = lambda topic, _bytes: self.raw_config[k](_bytes)
        else:
          self.kds = self.raw_config[k].deserialize
        self.config[k] = self.kds
      elif (k == 'value_deserializer'):
        if not (self.raw_config[k] is None) and (not hasattr(self.raw_config[k], 'deserialize') or not inspect.isroutine(self.raw_config[k].deserialize)):
          self.vds = lambda topic, _bytes: self.raw_config[k](_bytes)
        else:
          self.vds = self.raw_config[k].deserialize
        self.config[k] = self.vds
      elif (k in self.CONFIG_MAP.keys()):
        if not (self.raw_config[k] is None):
          self.cf_config[self.CONFIG_MAP[k]] = self.raw_config[k]
      elif (k in self.CONFIG_MAP_LOCAL):
        self.config[k] = self.raw_config[k]
      elif (k in self.CONFIG_MAP_NULL):
        self._log.warning("Warning: Unsupported kafka-python parameter passed to confluent wrapper: %s, %s", k, self.raw_config[k])
      else:
        self.cf_config[k] = self.raw_config[k]
    for oldk,newk in self.CONFIG_MAP.items():
      if newk in self.cf_config.keys():
        self.config[oldk] = self.cf_config[newk]
    # Remove empty parameters
    for k in list(self.cf_config.keys()):
      if (self.cf_config[k] is None) or (isinstance(self.cf_config[k], (str,)) and len(self.cf_config[k]) == 0):
        self._log.warning("Warning: Removing empty parameter %s passed to confluent wrapper", k)
        self.cf_config.pop(k)
    super().__init__(self.cf_config)
    if topics:
      self.subscribe(topics)

  def close(self):
    # confluent-kafka consumer causes crashes on close,
    # so this commits offsets and unsubscribes
    self.commit()
    self.unsubscribe()
    pass

  def _commit(self, offsets=None, asynchronous=False):
    self._log.debug("Executing Consumer commit.")
    try:
      if offsets==None:
        rv = super().commit(asynchronous=asynchronous)
      else:
        offs = []
        for k in offsets.keys():
          offs.append(TopicPartitionOffset(k.topic,k.partition,offsets[k].offset))
        rv = super().commit(offsets=offs,asynchronous=asynchronous)
    except KafkaException as ke:
      try:
        if ke.args[0].code() in [KafkaError._NO_OFFSET]:
          # ignore errors about committing offsets (means we have subscribed but not yet been assigned a particular topicpartition)
          # by doing this asynchronously. This matches the implicit kafka-python contract of not committing on topics
          # subscribed but not yet assigned.
          if offsets==None:
            return super().commit(asynchronous=True)
          else:
            return super().commit(offsets=offs,asynchronous=True)
        else:
          raise ke
      except:
        raise ke
    return rv

  def commit(self, offsets=None):
    return self._commit(offsets=offsets, asynchronous=False)

  def commit_async(self, offsets=None):
    # Right now callback is not implemented, so by leaving out of above we generate an exception
    # when a caller tries to use that functionality.
    return self._commit(offsets=offsets, asynchronous=True)

  def subscribe(self, topics=None, pattern=None, listener=None, **kwargs):
    self._log.info("Executing Consumer subscribe with topics=%s, pattern=%s, listener=%s", str(topics), str(pattern), str(listener))
    if not topics is None:
      topics = copy.copy(topics)
    else:
      topics = []
    if not pattern is None:
      if pattern[0] != '^':
        topics.append('^' + pattern)
      else:
        topics.append(pattern)
    if listener is not None:
      # the lambda expressions pass tuples with topic, partition, and offset (instead of just topic and partition),
      # but that is compatible with the intended API of ConsumerRebalanceListener (see https://github.com/dpkp/kafka-python/blob/f19e4238fb47ae2619f18731f0e0e9a3762cfa11/kafka/consumer/subscription_state.py)
      return super().subscribe(topics, on_assign=lambda c,tps: listener.on_partitions_assigned(tps), on_revoke=lambda c,tps: listener.on_partitions_revoked(tps), **kwargs)
    else:
      return super().subscribe(topics, **kwargs)

  def poll(self, timeout_ms=0, max_records=None):
    rvm = {}
    if (max_records is None) or max_records <= 0 or max_records > self.config['max_poll_records']:
      max_records = self.config['max_poll_records']
    self._log.debug("Entering Consumer poll with timeout_ms=%s and max_records=%s", str(timeout_ms), str(max_records))
    msgs = super().consume(max_records,timeout_ms/1000.0)
    self._log.debug("Consumer poll consume complete.")
    for msg in msgs:
      if not (msg is None) and msg.error() is None:
        rvk = TopicPartition(msg.topic(),msg.partition())
        rv = Message(rvk.topic, rvk.partition, msg.offset(), msg.timestamp(), msg.headers(), self.kds(rvk.topic,msg.key()), self.vds(rvk.topic,msg.value()))
        if rvk in rvm.keys():
          rvm[rvk].append(rv)
        else:
          rvm[rvk] = [rv]
    return rvm

  def assign(self, partitions):
    self._log.info("Executing Consumer assign with partitions=%s.", str(partitions))
    tps = []
    for tp in partitions:
      tps.append(TopicPartitionOffset(tp.topic, tp.partition))
    return super().assign(tps)

  def assign_and_seek(self, partoffs):
    self._log.info("Executing Consumer assign_and_seek with partoffs=%s.", str(partoffs))
    tps = []
    for tpo in partoffs:
      if (tpo.offset > 0) or (tpo.offset == OFFSET_BEGINNING) or (tpo.offset == OFFSET_END):
        tps.append(TopicPartitionOffset(tpo.topic, tpo.partition, tpo.offset))
      else:
        tps.append(TopicPartitionOffset(tpo.topic, tpo.partition))
    return super().assign(tps)

  def seek(self, tp, offset):
    self._log.debug("Executing Consumer seek.")
    return super().seek(TopicPartitionOffset(tp.topic, tp.partition, offset))

  def seek_to_beginning(self, *tps):
    self._log.debug("Executing Consumer seek_to_beginning.")
    if len(tps) < 1:
      tps = self.assignment()
    elif len(tps) == 1 and not isinstance(tps[0], TopicPartition):
      tps = tps[0] # confluent_kafka all TPs as a single passed list convention
    for tp in tps:
      try:
        self.seek(tp, OFFSET_BEGINNING)
      except:
        pass

  def seek_to_end(self, tps=None):
    self._log.debug("Executing Consumer seek_to_end.")
    if len(tps) < 1:
      tps = self.assignment()
    elif len(tps) == 1 and not isinstance(tps[0], TopicPartition):
      tps = tps[0] # confluent_kafka all TPs as a single passed list convention
    for tp in tps:
      try:
        self.seek(tp, OFFSET_END)
      except:
        pass

  def __iter__(self):
    return self

  def __next__(self):
    rv = self.poll(timeout_ms=-1000.0,max_records=1)
    if len(rv) == 1:
      return list(rv.values())[0][0]
    raise StopIteration("Poll Failed!")

class KafkaProducer(Producer):
  """
  Thin wrapper around confluent-kafka providing methods unique to kafka-python
  to allow them to be used interchangeably. Unless otherwise noted, parameters should
  be supplied as expected by the underlying confluent_kafka classes. Use at your own risk!
  """
  CONFIG_MAP = { 'ssl_cafile': 'ssl.ca.location',
                 'ssl_certfile': 'ssl.certificate.location',
                 'ssl_keyfile': 'ssl.key.location',
                 'ssl_crlfile': 'ssl.crl.location',
                 'ssl_password': 'ssl.key.password',
                 'ssl_ciphers': 'ssl.cipher.suites',
                 'max_request_size': 'message.max.bytes', # this maps the hard limit functionality appropriately
                 'bootstrap_servers': 'bootstrap.servers',
                 'client_id': 'client.id',
                 'group_id': 'group.id',
                 'acks': 'request.required.acks',
                 'compression_type': 'compression.type',
                 'retries': 'message.send.max.retries',
                 'linger_ms': 'queue.buffering.max.ms',
                 'request_timeout_ms': 'request.timeout.ms',
                 'retry_backoff_ms': 'retry.backoff.ms',
                 'reconnect_backoff_ms': 'reconnect.backoff.ms',
                 'reconnect_backoff_max_ms': 'reconnect.backoff.max.ms',
                 'max_in_flight_requests_per_connection': 'max.in.flight.requests.per.connection',
                 'metadata_max_age_ms': 'metadata.max.age.ms',
                 'receive_buffer_bytes': 'socket.receive.buffer.bytes',
                 'send_buffer_bytes': 'socket.send.buffer.bytes',
                 'security_protocol': 'security.protocol',
                 'sasl_mechanism': 'sasl.mechanism',
                 'sasl_plain_username': 'sasl.username',
                 'sasl_plain_password': 'sasl.password',
                 'sasl_kerberos_service_name': 'sasl.kerberos.service.name',
               }
  CONFIG_MAP_LOCAL = { 'produce_timeout': 30, # in s
                     }
  CONFIG_MAP_NULL = [ 'ssl_context',
                      'socket_options',
                      'batch_size',
                      'partitioner',
                      'buffer_memory',
                      'connections_max_idle_ms',
                      'max_block_ms',
                      'ssl_check_hostname',
                      'api_version',
                      'api_version_auto_timeout_ms',
                      'metrics_reporters',
                      'metrics_num_samples',
                      'metrics_sample_window_ms',
                      'selector',
                      'sasl_kerberos_domain_name',
                      'sasl_oauth_token_provider',
                    ]

  enable_flush_workaround = False

  def __init__(self, **configs):
    self._log = logging.getLogger(__name__)
    self.raw_config = configs
    self.cf_config = {}
    self.config = {}
    self.ks = lambda topic, _bytes: bytes(_bytes)
    self.vs = lambda topic, _bytes: bytes(_bytes)
    for k in self.CONFIG_MAP_LOCAL.keys():
      self.config[k] = self.CONFIG_MAP_LOCAL[k]
    for k in self.raw_config.keys():
      # adjust default produce_timeout if not separately set, and
      # other parameters are set that require it to adjust.
      if (k in ['delivery.timeout.ms', 'message.timeout.ms', 'transaction.timeout.ms']) and ('produce_timeout' not in self.raw_config.keys()):
        self.config['produce_timeout'] = self.raw_config[k]/1000.0
      if (k == 'key_serializer'):
        if not (self.raw_config[k] is None) and (not hasattr(self.raw_config[k], 'serialize') or not inspect.isroutine(self.raw_config[k].serialize)):
          self.ks = lambda topic, _bytes: self.raw_config[k](_bytes)
        else:
          self.ks = self.raw_config[k].serialize
        self.config[k] = self.ks
      elif (k == 'value_serializer'):
        if not (self.raw_config[k] is None) and (not hasattr(self.raw_config[k], 'serialize') or not inspect.isroutine(self.raw_config[k].serialize)):
          self.vs = lambda topic, _bytes: self.raw_config[k](_bytes)
        else:
          self.vs = self.raw_config[k].serialize
        self.config[k] = self.vs
      elif (k in self.CONFIG_MAP.keys()):
        if not (self.raw_config[k] is None):
          self.cf_config[self.CONFIG_MAP[k]] = self.raw_config[k]
      elif (k in self.CONFIG_MAP_LOCAL):
        self.config[k] = self.raw_config[k]
      elif (k in self.CONFIG_MAP_NULL):
        self._log.warning("Warning: Unsupported kafka-python parameter passed to confluent wrapper: %s, %s", k, self.raw_config[k])
      else:
        self.cf_config[k] = self.raw_config[k]
    for oldk,newk in self.CONFIG_MAP.items():
      if newk in self.cf_config.keys():
        self.config[oldk] = self.cf_config[newk]
    # Remove empty parameters
    for k in list(self.cf_config.keys()):
      if (self.cf_config[k] is None) or (isinstance(self.cf_config[k], (str,)) and len(self.cf_config[k]) == 0):
        self._log.warning("Warning: Removing empty parameter %s passed to confluent wrapper", k)
        self.cf_config.pop(k)
    self.messages_processed = 0
    self.messages_processed_lock = Lock()
    super().__init__(self.cf_config)
    if KafkaProducer.enable_flush_workaround:
      self.flush = self._flush_workaround
    else:
      self.flush = self._flush_native

  def close(self):
    # confluent-kafka has no concept of a "close" operation,
    # so this flushes queues only
    self.flush()
    pass

  def _flush_native(self, timeout="default", timeout_jiffy=0.1):
    if timeout is None:
      # confluent_kafka uses -1 for infinite timeout
      timeout = -1
    elif timeout == "default":
      timeout = self.config['produce_timeout']
    return super().flush(timeout)

  def _flush_workaround(self, timeout="default", timeout_jiffy=0.1):
    # librdkafka 1.8.0 changed the behavior of flush() to ignore linger_ms and
    # immediately attempt to send messages. Unfortunately, that change added
    # a call to rd_kafka_all_brokers_wakeup , which seems to cause a hang of
    # some type (perhaps a deadlock given the warning in kafka_all_brokers_wakeup
    # about making sure certain locks are not held?).
    #
    # This eventually results in non-sensical exceptions being thrown. We
    # fix it here by implementing flush as sucessive polling directly.
    #
    # timeout = None is infinite timeout
    # timeout = "default" (or any non-number) should never be passed by callers,
    #           but is used here to indicate use of configured default.
    #
    if timeout is None:
      # confluent_kafka uses -1 for infinite timeout
      timeout = -1
    elif timeout == "default":
      timeout = self.config['produce_timeout']
    with self.messages_processed_lock:
      left = len(self)
      final_processed = self.messages_processed+left
    initial = left
    self._log.debug("Entering Producer flush with timeout=%s and left=%s.", str(timeout), str(left))
    if timeout<0:
      while left > 0:
        con = self.poll(timeout_jiffy)
        initial -= con
        with self.messages_processed_lock:
          left = min([len(self),initial,max([0,final_processed-self.messages_processed])])
        self._log.debug("Producer flush poll cycle complete, left=%s.", str(left))
    elif left > 0:
      con = self.poll(timeout)
      initial -= con
      with self.messages_processed_lock:
        left = min([len(self),initial,max([0,final_processed-self.messages_processed])])
      self._log.debug("Producer flush single poll cycle complete, left=%s.", str(left))
    if max([left,0]) != 0:
      self._log.info("Producer flush complete, but messages left=%s greater than zero.", str(max([left,0])))
    else:
      self._log.debug("Producer flush complete.")
    return max([left,0])

  def poll(self, timeout=0):
    # timeout = 0 means nowait (default)
    # timeout = None means infinite timeout
    self._log.debug("Executing Producer poll with timeout=%s.", str(timeout))
    if timeout is None:
      # confluent_kafka uses -1 for infinite timeout
      timeout = -1
    rv = super().poll(timeout)
    if rv > 0:
      with self.messages_processed_lock:
        self.messages_processed += rv
    return rv

  def send(self, topic, value=None, key=None, headers=None, partition=0, timestamp_ms=None):
    self._log.debug("Executing Producer send to topic=%s, with value=%s, key=%s, headers=%s, partition=%s, timestamp_ms=%s.", str(topic), str(value), str(key), str(headers), str(partition), str(timestamp_ms))
    if key!=None:
      key = self.ks(topic, key)
    if value!=None:
      value = self.vs(topic, value)
    rv = FutureRecordMetadata(self, len(value) if value!=None else -1, len(key) if key!=None else -1)
    # wish this were simpler, but the underlying library doesn't like None when no value should be passed.
    if key is None:
      key = b''
    if value is None:
      value = b''
    if not (headers is None):
      self.produce(topic, value, key, partition, rv.base_callback, timestamp_ms if timestamp_ms!=None else int(time()*1000), headers)
    elif not (timestamp_ms is None):
      self.produce(topic, value, key, partition, rv.base_callback, timestamp_ms)
    else:
      self.produce(topic, value, key, partition, rv.base_callback)
    return rv



from threading import Thread
import inspect
from kafkacrypto.utils import format_exception_shim
import pysodium
from time import time,sleep
import msgpack
import logging
from kafkacrypto import TopicPartition,TopicPartitionOffset,OFFSET_BEGINNING
from kafkacrypto.common_wrapper import AbstractSerializer, AbstractDeserializer
from kafkacrypto.base import KafkaCryptoBase
from kafkacrypto.exceptions import KafkaCryptoError, KafkaCryptoSerializeError
from kafkacrypto.message import KafkaCryptoMessage
from kafkacrypto.ratchet import Ratchet
from kafkacrypto.keygenerator import KeyGenerator
from kafkacrypto.utils import log_limited, msgpack_default_pack

class KafkaCrypto(KafkaCryptoBase):
  """Class handling the sending and receiving of encrypted messages.

  Keyword Arguments:
              nodeID (str): Node ID
        kp (KafkaProducer): Pre-initialized KafkaProducer, ready for
                            handling crypto-keying messages. Should
                            not be used elsewhere, as this class
                            changes some configuration values. Always
                            closed on close.
        kc (KafkaConsumer): Pre-initialized KafkaConsumer, ready for
       	       	       	    handling crypto-keying messages. Should
                            not be used elsewhere, as this class
                            changes some configuration values. Always
                            closed on close.
         config (str,file): Filename or File IO object in which
                            configuration data is stored. Set to None
                            to load from the default location based
                            on nodeID. Must be seekable, with read/
                            write permission, honor sync requests,
                            and not be written by any other program.
                            Only closed if opened by us.
           cryptokey (obj): Optional object implementing the
                            necessary public/private key functions
                            (get/sign_spk,get/use_epks,
                            wrap/unwrap_opaque).
                            Set to None to load from the default
                            location in the configuration file.
                seed (obj): Optional object implementing the
                            necessary ratchet functions
                            (increment, generate). Set to None to
                            load from the default location in
                            the configuration file. Only closed
                            if opened by us.
  """

  #
  # Managed by our process, and protected by _lock since they are also
  # utilized by SerDes subclasses:
  # _tps_updated: set to true anytime a non-management process updates
  #               the list of topic-partitions.
  #         _tps: dict of topic-partitions currently listening to. Needed because
  #               KafkaConsumer does not support incremental assignments, so every
  #               change requires (re)subscribing to all TopicPartitions.
  #               indexed by full topic.
  # _tps_offsets: dict of offsets for each tps. Used since kp,kc may have been
  #               configured without groupIDs, so we have to keep track of our
  #               own offsets. This is the next offset to process.
  #               indexed by full topic.
  # _subs_needed: array of root topics for which new subscriptions are needed.
  #   _subs_last: dict of subscriptions that were requested (by root topic).
  #               Each one is itself an array with the timestamp, and an
  #               array of key indices requested.
  #   _cur_pgens: dict of keyidx's for the current producer generator, indexed
  #               by root topic.
  #       _pgens: dict of production encryption key generators, indexed by root
  #               topic, then by 'keyidx', and 'key'/'value'/'secret'.
  #   _new_pgens: Set to true when new pgens were added
  #      _cwaits: dict of statuses on waiting for different key indices indexed by 
  #               root topic and then key index. Used to implement the
  #               wait once logic to ensure no undecryptable messages at
  #               transient key change events.
  #       _cgens: dict of consumption encryption key generators, indexed by root
  #               topic. Each topic has a list of available generators, indexed
  #               by 'keyidx', and finally 'key'/'value'/'secret'.
  #

  def __init__(self, nodeID, kp, kc, config=None, cryptokey=None, seed=None):
    super().__init__(nodeID, kp, kc, config, cryptokey)
    if (seed is None):
      seed = self._cryptostore.load_value('ratchet')
      if seed!=None and seed.startswith('file#'):
        seed = seed[5:]
      if (seed is None):
        seed = self._nodeID + ".seed"
        self._cryptostore.store_value('ratchet', 'file#' + seed)
    self._seed_close = False
    if (isinstance(seed,(str,))):
      self._seed_close = True
      seed = Ratchet(file=seed)
    if (not hasattr(seed, 'increment') or not inspect.isroutine(seed.increment) or not hasattr(seed, 'get_key_value_generators') or not inspect.isroutine(seed.get_key_value_generators)):
      raise KafkaCryptoError("Invalid seed source supplied!")

    self._seed = seed
    self._tps_updated = False
    self._tps = {}
    self._tps_offsets = {}
    self._cur_pgens = {}
    self._pgens = {}
    self._new_pgens = False
    self._cwaits = {}
    self._cgens = {}
    self._subs_needed = []
    self._subs_last = {}
    self._last_time = time()

    # Add management frame TPS
    for ntp in [self.MGMT_TOPIC_CHAINS, self.MGMT_TOPIC_ALLOWLIST, self.MGMT_TOPIC_DENYLIST]:
      self._tps[ntp] = TopicPartition(ntp,0)
      self._tps_offsets[ntp] = OFFSET_BEGINNING
      self._tps_updated = True

    kvs = self._cryptostore.load_opaque_value('oldkeys',section="crypto")
    if not (kvs is None):
      kvs = msgpack.unpackb(kvs,raw=True)
      if 'pgens' in kvs.keys() or b'pgens' in kvs.keys():
        self._logger.info("Found pgens to load.")
        t = 'pgens'
        if not (t in kvs.keys()):
          t = b'pgens'
        for [root,ki,ks,b] in kvs[t]:
          if isinstance(root,(bytes,bytearray)):
            self._logger.debug("loaded a pgen root in bytes (should be string)")
            root = root.decode('utf-8')
          self._logger.info("Attempting load of root=%s, ki=%s, b=%s at %s", root, ki, b, time())
          if time()-b < self.CRYPTO_MAX_PGEN_AGE:
            if not (root in self._pgens.keys()):
              self._pgens[root] = {}
            self._logger.info("Loaded old key keyindex=%s, birth=%s for root=%s", ki, b, root)
            self._pgens[root][ki] = {}
            self._pgens[root][ki]['birth'] = b
            self._pgens[root][ki]['secret'] = ks
            # stored pgens do not have generators as they should never be used for active production
            # (but secret stays around so lost consumers can catch up)

    self._mgmt_thread_stop = False
    self._mgmt_thread = Thread(target=self._process_mgmt_messages,daemon=True)
    self._mgmt_thread.start()

  def close(self):
    self._mgmt_thread_stop = True
    try:
      self._mgmt_thread.join()
    except:
      pass
    finally:
      self._mgmt_thread = None
    try:
      if self._seed_close:
        self._seed.close()
    except:
      pass
    finally:
      self._seed = None
    super().close()

  # Main background processing loop. Must assume that it can "die" at any
  # time, even mid-stride, so ordering of operations to ensure atomicity and
  # durability is critical.
  def _process_mgmt_messages(self):
    while True:
      if self._mgmt_thread_stop:
        break
      # First, process messages
      # we are the only thread ever using _kc, _kp, so we do not need the lock to use them
      msgs = self._kc.poll(timeout_ms=self.MGMT_POLL_INTERVAL, max_records=self.MGMT_POLL_RECORDS)
      # but to actually process messages, we need the lock
      for tp,msgset in msgs.items():
        self._lock.acquire()
        for msg in msgset:
          topic = msg.topic
          if (isinstance(topic,(bytes,bytearray))):
            self._logger.debug("passed a topic in bytes (should be string)")
            topic = topic.decode('utf-8')
          self._tps_offsets[topic] = msg.offset+1
          self._logger.debug("Processing message: %s", msg)
          if topic[-len(self.TOPIC_SUFFIX_REQS):] == self.TOPIC_SUFFIX_REQS:
            root = topic[:-len(self.TOPIC_SUFFIX_REQS)]
            # A new receiver: send all requested keys
            try:
              kreq = msgpack.unpackb(msg.key,raw=True)
              if root in self._pgens.keys():
                ki = []
                s = []
                for ski,sk in self._pgens[root].items():
                  if ski in kreq:
                    ki.append(ski)
                    s.append(sk['secret'])
                if len(ki) > 0:
                  k = msgpack.packb(ki, default=msgpack_default_pack, use_bin_type=True)
                  v = self._cryptoexchange.encrypt_keys(ki, s, root, msgval=msg.value)
                  if not (v is None):
                    self._logger.info("Sending current encryption keys for root=%s to new receiver, msgkey=%s.", root, k)
                    self._kp.send(root + self.TOPIC_SUFFIX_KEYS, key=k, value=v)
                  else:
                    self._logger.info("Failed sending current encryption keys for root=%s to new receiver.", root)
                else:
                  self._logger.info("No keys for root=%s to send to new receiver.", root)
            except Exception as e:
              self._parent._logger.warning("".join(format_exception_shim(e)))
          elif topic[-len(self.TOPIC_SUFFIX_KEYS):] == self.TOPIC_SUFFIX_KEYS:
            root = topic[:-len(self.TOPIC_SUFFIX_KEYS)]
            # New key(s)
            nks = self._cryptoexchange.decrypt_keys(root,msgval=msg.value)
            if not (nks is None):
              for nki,nk in nks.items():
                self._logger.info("Received new encryption key for root=%s, key index=%s, msgkey=%s", root, nki, msg.key)
                # do not clopper other keys that may exist
                if not (root in self._cgens.keys()):
                  self._cgens[root] = {}
                # but do clobber the same topic,keyID entry
                self._cgens[root][nki] = {}
                self._cgens[root][nki]['key'], self._cgens[root][nki]['value'] = KeyGenerator.get_key_value_generators(nk)
                self._cgens[root][nki]['secret'] = nk
                # now that we have this key index, clear from request lists
                if root in self._cwaits.keys():
                  self._cwaits[root].pop(nki, None)
              if root in self._subs_last:
                eki = set(self._subs_last[root][1])
                eki.difference_update(set(nks.keys()))
                if len(eki) > 0:
                  self._logger.warning("For root=%s, the keys %s were requested but not received.", root, eki)
                self._subs_last.pop(root,None)
          elif topic == self.MGMT_TOPIC_CHAINS:
            # New candidate public key chain
            self._logger.info("Received new chain message: %s", msg)
            if msg.key == bytes(self._cryptokey.get_spk()):
              self._logger.debug("Key matches ours. Validating Chain.")
              newchain = self._cryptoexchange.replace_spk_chain(msg.value)
              if not (newchain is None):
                self._logger.info("New chain is superior, using it.")
                self._cryptostore.store_value('chain',newchain,section='crypto')
          elif topic == self.MGMT_TOPIC_ALLOWLIST:
            self._logger.info("Received new allowlist message: %s", msg)
            allow = self._cryptoexchange.add_allowlist(msg.value)
            if not (allow is None):
              c = pysodium.crypto_hash_sha256(allow)
              self._cryptostore.store_value(c,allow,section='allowlist')
          elif topic == self.MGMT_TOPIC_DENYLIST:
            self._logger.info("Received new denylist message: %s", msg)
            deny = self._cryptoexchange.add_denylist(msg.value)
            if not (deny is None):
              c = pysodium.crypto_hash_sha256(deny)
       	      self._cryptostore.store_value(c,deny,section='denylist')
          else:
            # unknown object
            log_limited(self._logger.warning, "Unknown topic type in message: %s", msg)
        self._lock.release()

      # Flush producer
      try:
        self._kp.flush(timeout=self.MGMT_FLUSH_TIME)
      except Exception as e:
        self._parent._logger.warning("".join(format_exception_shim(e)))

      # Second, deal with subscription changes
      self._lock.acquire()
      self._logger.debug("Processing subscription changes.")
      if self._tps_updated == True:
        self._logger.debug("Subscriptions changed, adjusting.")
        tpo = []
        for tk in self._tps.keys():
          tpo.append(TopicPartitionOffset(self._tps[tk].topic, self._tps[tk].partition, self._tps_offsets[tk]))
        self._kc.assign_and_seek(tpo)
        self._logger.debug("Subscriptions adjusted.")
        if (self._kc.config['group_id'] is None):
          self._logger.info("No group_id, seeking to beginning.")
          self._kc.seek_to_beginning()
        for tk in self._tps.keys():
          if (tk[-len(self.TOPIC_SUFFIX_KEYS):] == self.TOPIC_SUFFIX_KEYS):
            root = tk[:-len(self.TOPIC_SUFFIX_KEYS)]
            if not (root in self._subs_needed):
              self._subs_needed.append(root)
        self._tps_updated = False
      self._lock.release()

      # Third, deal with topics needing subscriptions
      self._lock.acquire()
      self._logger.debug("Process subscriptions.")
      subs_needed_next = []
      for root in self._subs_needed:
        if root in self._cwaits.keys():
          self._logger.debug("Attempting (Re)subscribe to root=%s", root)
          kis = list(self._cwaits[root].keys())
          if len(kis) > 0:
            if (not (root in self._subs_last.keys()) or self._subs_last[root][0]+self.CRYPTO_SUB_INTERVAL<time()):
              k = msgpack.packb(kis, default=msgpack_default_pack, use_bin_type=True)
              v2 = self._cryptoexchange.signed_epks(root)
              if not (k is None) and not (v2 is None):
                self._logger.info("Sending new subscribe request(s) for root=%s, msgkey=%s", root, k)
                for v in v2:
                  self._kp.send(root + self.TOPIC_SUFFIX_SUBS, key=k, value=v)
                  if self._cryptoexchange.direct_request_spk_chain():
                    # if it may succeed, send directly as well
                    self._kp.send(root + self.TOPIC_SUFFIX_REQS, key=k, value=v)
                self._subs_last[root] = [time(),kis]
              else:
                self._logger.info("Failed to send new subscribe request for root=%s", root)
                subs_needed_next.append(root)
            else:
              self._logger.debug("Deferring (re)subscriptions for root=%s due to pending key request.", root)
              subs_needed_next.append(root)
          else:
            self._logger.debug("No new keys needed for root=%s", root)
      self._subs_needed = subs_needed_next
      self._lock.release()

      # Flush producer
      try:
        self._kp.flush(timeout=self.MGMT_FLUSH_TIME)
      except Exception as e:
        self._parent._logger.warning("".join(format_exception_shim(e)))

      # Fourth, periodically increment ratchet and prune old keys
      self._logger.debug("Checking ratchet time.")
      self._lock.acquire()
      if self._last_time+self.CRYPTO_RATCHET_INTERVAL < time():
        self._logger.info("Periodic ratchet increment.")
        self._last_time = time()
        # prune
        for root,pgens in self._pgens.items():
          rki = []
          for ki,kv in pgens.items():
            if (kv['birth']+self.CRYPTO_MAX_PGEN_AGE<time()):
              rki.append(ki)
          for ki in rki:
            self._pgens[root].pop(ki)
        # increment
        self._cur_pgens = {}
        self._seed.increment()
      self._lock.release()

      # Fifth, write new producer keys if requested
      self._logger.debug("Checking producer keys.")
      self._lock.acquire()
      if self._new_pgens:
        self._logger.info("(Re)writing producer keys.")
        self._new_pgens = False
        kvs = {}
        kvs['pgens'] = []
        for root,pgens in self._pgens.items():
          for ki,kv in pgens.items():
            kvs['pgens'].append([root,ki,kv['secret'],kv['birth']])
            # stored pgens do not have generators as they should never be used for active production
            # (but secret stays around so lost consumers can catch up)
        self._logger.info("Saving %s old production keys.", len(kvs['pgens']))
        self._cryptostore.store_opaque_value('oldkeys',msgpack.packb(kvs, default=msgpack_default_pack, use_bin_type=True),section="crypto")
      self._lock.release()
  
      # Finally, loop back to poll again
  # end of __process_mgmt_messages

  #
  # We have two internal subclasses that implement the Kafka SerDes interface,
  # and handle new keys/subscriptions as needed.
  #
  class Serializer(AbstractSerializer):
    def __init__(self, parent, kv):
      self._kv = kv
      self._parent = parent

    def serialize(self, topic, value):
      if (isinstance(topic,(bytes,bytearray))):
        self._parent._logger.debug("passed a topic in bytes (should be string)")
        topic = topic.decode('utf-8')
      if value is None:
        return None
      if (isinstance(value,(KafkaCryptoMessage,))):
        if (value.isCleartext(retry=False)):
          value = value.getMessage(retry=False)
        else:
          # already serialized and encrypted
          return bytes(value)
      if (not isinstance(value,(bytes,bytearray))):
        raise KafkaCryptoSerializeError("Passed value is not bytes or a KafkaCryptoMessage")
      root = self._parent.get_root(topic)
      self._parent._lock.acquire()
      try:
        #
        # Slow path: if we don't already have a generator for this topic,
        # we have to make one.
        #
        if not (root in self._parent._cur_pgens.keys()):
          self._parent.get_producer(root)
        # Use generator (new or existing)
        keyidx = self._parent._cur_pgens[root]
        pgen = self._parent._pgens[root][keyidx]
        gen = pgen[self._kv]
        salt = gen.salt()
        key,nonce = gen.generate()
        msg = b'\x01' + msgpack.packb([keyidx,salt,pysodium.crypto_secretbox(value,nonce,key)], default=msgpack_default_pack, use_bin_type=True)
      except Exception as e:
        self._parent._logger.warning("".join(format_exception_shim(e)))
      finally:
        self._parent._lock.release()
      return msg

  class Deserializer(AbstractDeserializer):
    def	__init__(self, parent, kv, max_key_wait_intervals, key_wait_interval):
      self._kv = kv
      self._parent = parent
      self.MAX_WAIT_INTERVALS = max_key_wait_intervals
      self.WAIT_INTERVAL = key_wait_interval

    def deserialize(self, topic, bytes_):
      if (isinstance(topic,(bytes,bytearray))):
        self._parent._logger.debug("passed a topic in bytes (should be string)")
        topic = topic.decode('utf-8')
      if bytes_ is None:
        return None
      root = self._parent.get_root(topic)
      if len(bytes_) < 1 or bytes_[0] != 1:
        # Not a KafkaCrypto format message
        return KafkaCryptoMessage.fromEncryptedBytes(bytes_,deser=self,topic=topic,decryptable=False)
      try:
        msg = msgpack.unpackb(bytes_[1:],raw=True)
        if (len(msg) != 3):
          raise KafkaCryptoSerializeError("Malformed Message!")
      except Exception as e:
        # Not a KafkaCrypto format message
        self._parent._logger.debug("".join(format_exception_shim(e)))
        return KafkaCryptoMessage.fromEncryptedBytes(bytes_,deser=self,topic=topic,decryptable=False)
      ki = msg[0]
      salt = msg[1]
      msg = msg[2]
      self._parent._lock.acquire()
      i = 1
      initial_waiter = False
      try:
        while ((not (root in self._parent._cgens.keys()) or not (ki in self._parent._cgens[root].keys())) and i>0):
          if (not (root in self._parent._cwaits.keys()) or not (ki in self._parent._cwaits[root].keys())):
            self._parent._logger.debug("Setting initial wait for root=%s, key index=%s.", root, ki)
            initial_waiter = True
            # first time we see a topic/key index pair, we wait the initial interval for key exchange
            i = self._parent.DESER_INITIAL_WAIT_INTERVALS
            if not (root in self._parent._cwaits.keys()):
              self._parent._cwaits[root] = {}
            self._parent._cwaits[root][ki] = i
          elif initial_waiter:
            self._parent._cwaits[root][ki] -= 1
          elif (root in self._parent._cwaits.keys()) and (ki in self._parent._cwaits[root].keys()):
            i = self._parent._cwaits[root][ki]
          else:
            i = self.MAX_WAIT_INTERVALS
       	  ntp =	(root+self._parent.TOPIC_SUFFIX_KEYS)
       	  if not (ntp in self._parent._tps):
       	    self._parent._tps[ntp] = TopicPartition(ntp,0)
            self._parent._tps_offsets[ntp] = 0
            self._parent._tps_updated = True
          else:
            if not (root in self._parent._subs_needed):
              self._parent._subs_needed.append(root)
          i=i-1
          if (i > 0):
            self._parent._lock.release()
            sleep(self.WAIT_INTERVAL)
            self._parent._lock.acquire()
        if not (root in self._parent._cgens.keys()) or not (ki in self._parent._cgens[root].keys()):
          self._parent._logger.debug("No decryption key found for root=%s, key index=%s. Returning encrypted message.", root, ki)
          raise ValueError
        gen = self._parent._cgens[root][ki][self._kv]
        key,nonce = gen.generate(salt=salt)
        msg = KafkaCryptoMessage(pysodium.crypto_secretbox_open(msg,nonce,key),ipt=True)
      except:
        # errors here mean the requisite key isnt (yet) available
        return KafkaCryptoMessage.fromEncryptedBytes(bytes_,deser=self,topic=topic,decryptable=True)
      finally:
        self._parent._lock.release()
      return msg

  def getKeySerializer(self):
    return self.Serializer(self,'key')
  def getValueSerializer(self):
    return self.Serializer(self,'value')
  def getKeyDeserializer(self, max_key_wait_intervals=0, key_wait_interval=1):
    return self.Deserializer(self,'key',max_key_wait_intervals,key_wait_interval)
  def getValueDeserializer(self, max_key_wait_intervals=0, key_wait_interval=1):
    return self.Deserializer(self,'value',max_key_wait_intervals,key_wait_interval)

  def get_producer(self,root):
    if isinstance(root,(bytes,bytearray)):
      self._logger.debug("passed a root in bytes (should be string)")
      root = root.decode('utf-8')
    if (self.MGMT_LONG_KEYINDEX == True):
      ki,ks,kg,vg = self._seed.get_key_value_generators(root, node=self._cryptokey.get_spk())
    else:
      ki,ks,kg,vg = self._seed.get_key_value_generators(root)
    self._cur_pgens[root] = ki
    if not (root in self._pgens):
      self._pgens[root] = {}
    if ki in self._pgens[root] and (not ('key' in self._pgens[root][ki].keys()) or not ('value' in self._pgens[root][ki].keys())):
      self._logger.critical("Potential key reuse scenario detected! Incrementing ratchet for safety.")
      self._seed.increment()
    elif not (ki in self._pgens[root].keys()):
      self._new_pgens = True
      self._pgens[root][ki] = {}
      self._pgens[root][ki]['key'] = kg
      self._pgens[root][ki]['value'] = vg
      self._pgens[root][ki]['secret'] = ks
      self._pgens[root][ki]['birth'] = time()
      ntp = (root+self.TOPIC_SUFFIX_REQS)
      if not (ntp in self._tps):
        self._tps[ntp] = TopicPartition(ntp,0)
        self._tps_offsets[ntp] = 0
        self._tps_updated = True
      self._logger.info("Got new producer key for root=%s. New key index=%s", root, ki)

